import cv2
import numpy as np
from PIL import Image, ImageDraw


class Isometry:
    '''class to get isometric image from three images in openCV/Pillow format.
    params:
        front_image;
        end_image;
        up_image;
        angle - angle to skew images to make isometry (default = 30 degrees);
        is_crop_white_background - if True: white background will be cropped from all input images (default = False);
        is_pil_format - if True: expected input images in Pillow format. Final isometry also will be in Pillow format (default=False)
    return:
        main function "get_isometric_image" returns final isomety image in openCV/numpy format
    '''
    def __init__(self, front_image, end_image, up_image, angle=30, is_crop_white_background=False, is_pil_format=False):
        self.is_pil_format = is_pil_format
        self.angle = angle
        if self.is_pil_format:
            front_image = Isometry.pil_to_cvopen(front_image)
            end_image = Isometry.pil_to_cvopen(end_image)
            up_image = Isometry.pil_to_cvopen(up_image)
        else:
            front_image = front_image
            end_image = end_image
            up_image = up_image
        if is_crop_white_background:
            front_image = Isometry.crop_white_background(front_image)
            end_image = Isometry.crop_white_background(end_image)
            up_image = Isometry.crop_white_background(up_image)
        self.front_image = front_image
        self.end_image = cv2.resize(end_image, (end_image.shape[1], front_image.shape[0]))
        self.up_image = cv2.resize(up_image, (front_image.shape[1], end_image.shape[1]))

    @staticmethod
    def pil_to_cvopen(image):
        image = image.convert('RGB')
        image_cv = np.array(image)
        image_cv = image_cv[:, :, ::-1].copy()
        return image_cv

    @staticmethod
    def cvopen_to_pil(image):
        image_cv = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image_pil = Image.fromarray(image_cv)
        return image_pil

    @staticmethod
    def crop_white_background(image):
        # Convert to gray, and threshold
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        _, threshed = cv2.threshold(gray, 240, 255, cv2.THRESH_BINARY_INV)
        # Morph-op to remove noise
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (11, 11))
        morphed = cv2.morphologyEx(threshed, cv2.MORPH_CLOSE, kernel)
        # Find the max-area contour
        cnts = cv2.findContours(morphed, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[-2]
        cnt = sorted(cnts, key=cv2.contourArea)[-1]
        x, y, w, h = cv2.boundingRect(cnt)
        dst = image[y:y + h, x:x + w]
        return dst

    @staticmethod
    def get_polygon_transparent_mask(image, coordinates):
        mask = Image.new("L", image.size, 0)
        draw = ImageDraw.Draw(mask)
        draw.polygon(coordinates, fill=255)
        return mask

    @staticmethod
    def get_blank_pil_image(width, height, factor=1):
        white_background = (255, 255, 255)
        image_size = (int(width * factor), int(height * factor))
        return Image.new('RGB', image_size, white_background)

    def __skew_image(self, image, side):
        height, width = image.shape[:2]
        pts_src = np.array(
            [[1, 1],
             [width, 1],
             [width, height],
             [1, height]], dtype=float) - 1
        new_width = int(width * np.cos(self.angle * np.pi / 180.))
        added_height = int(width * np.sin(self.angle * np.pi / 180.))
        new_height = height + added_height
        if side == 'end':
            pts_dst = np.array(
                [[1, 1],
                 [new_width, added_height],
                 [new_width, new_height],
                 [1, height]], dtype=float) - 1
            coordinates = [
                (0, 0),
                (0, height - 1),
                (new_width, new_height - 1),
                (new_width, added_height)
            ]
        elif side == 'front':
            pts_dst = np.array(
                [[1, added_height],
                 [new_width, 1],
                 [new_width, height],
                 [1, new_height]], dtype=float) - 1
            coordinates = [
                (0, added_height - 1),
                (0, new_height - 2),
                (new_width, height - 2),
                (new_width, 0)
            ]
        elif side == 'up':
            sin = np.sin(self.angle * np.pi / 180.)
            cos = np.cos(self.angle * np.pi / 180.)
            new_width = int(height * cos) + int(width * cos)
            new_height = int(height * sin) + int(width * sin)
            pts_dst = np.array(
                [[1, int(width * sin)],
                 [int(width * cos), 1],
                 [new_width, int(height * sin)],
                 [int(height * cos), new_height]], dtype=float) - 1
            coordinates = [
                (3, int(width * sin) - 2),
                (int(height * cos), new_height - 2),
                (new_width - 2, int(height * sin) - 1),
                (int(width * cos) - 1, 1)
            ]
        projected_img = np.zeros((new_height, new_width, 4), dtype=np.uint8)
        mask = np.zeros((new_height, new_width), dtype=np.uint8)
        cv2.fillConvexPoly(mask, pts_dst.astype(int), 255, 16)
        projected_img[:, :, 3] = mask
        h = cv2.findHomography(pts_src, pts_dst)[0]
        projected_img[:, :, :3] += cv2.warpPerspective(image, h, (new_width, new_height))
        projected_img_pil = Isometry.cvopen_to_pil(projected_img)
        return projected_img_pil, Isometry.get_polygon_transparent_mask(projected_img_pil, coordinates), coordinates

    def get_isometric_image(self):
        end_skew, end_mask, _ = self.__skew_image(self.end_image, 'end')
        front_skew, front_mask, _ = self.__skew_image(self.front_image, 'front')
        up_skew, up_mask, coordinates = self.__skew_image(self.up_image, 'up')
        blank_image = Isometry.get_blank_pil_image(up_skew.size[0], up_skew.size[1] + self.end_image.shape[0], 1.1)
        end_skew_offset = (0, coordinates[0][1] + 10)
        front_skew_offset = (
            end_skew_offset[0] + end_skew.size[0],
            end_skew_offset[1] + end_skew.size[1] - front_skew.size[1]
        )
        up_skew_offset = (
            end_skew_offset[0],
            end_skew_offset[1] - coordinates[0][1]
        )
        blank_image.paste(end_skew, end_skew_offset, end_mask)
        blank_image.paste(front_skew, front_skew_offset, front_mask)
        blank_image.paste(up_skew, up_skew_offset, up_mask)
        blank_image = Isometry.cvopen_to_pil(
            Isometry.crop_white_background(
                Isometry.pil_to_cvopen(blank_image)
            )
        )
        if self.is_pil_format:
            return blank_image
        else:
            return Isometry.pil_to_cvopen(blank_image)
